@@ -1,4 +1,4 @@
-/* Copyright (C) 1999-2004, Ghostgum Software Pty Ltd.  All rights reserved.
+/* Copyright (C) 1999-2003, Ghostgum Software Pty Ltd.  All rights reserved.
   
   This software is provided AS-IS with no warranty, either express or
   implied.
@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-// $Id: dwsetup.cpp,v 1.7.2.1 2004/01/30 09:08:42 ghostgum Exp $
+// $Id: dwsetup.cpp,v 1.8.2.1 2004/08/18 03:06:15 ghostgum Exp $
 //
 //
 // This is the setup program for Win32 Ghostscript
@@ -37,10 +37,10 @@
 // uninstall log files are to be placed.  
 // Subsequent lines contain files to be copied (but not directories).
 // For example, filelist.txt might contain:
-//   GPL Ghostscript 8.01
-//   gs8.01
-//   gs8.01\bin\gsdll32.dll
-//   gs8.01\lib\gs_init.ps
+//   GPL Ghostscript 8.15
+//   gs8.15
+//   gs8.15\bin\gsdll32.dll
+//   gs8.15\lib\gs_init.ps
 // The file fontlist.txt might contain:
 //   GPL Ghostscript Fonts
 //   fonts
@@ -750,7 +750,11 @@ install_prog()
 	strcat(szLIB, cinst.GetMainDir());
 	strcat(szLIB, "\\lib;");
 	strcat(szLIB, g_szTargetDir);
-	strcat(szLIB, "\\fonts");
+	strcat(szLIB, "\\fonts;");
+	strcat(szLIB, g_szTargetDir);
+	strcat(szLIB, "\\");
+	strcat(szLIB, cinst.GetMainDir());
+	strcat(szLIB, "\\Resource");
 	if (!cinst.UpdateRegistryValue(regkey1, regkey2, "GS_LIB", szLIB)) {
 		gs_addmess("Failed to add registry value\n");
 		return FALSE;
