@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-// $Id: dwuninst.cpp,v 1.4.6.1 2004/01/30 09:08:42 ghostgum Exp $
+// $Id: dwuninst.cpp,v 1.4.2.1 2004/08/18 03:06:15 ghostgum Exp $
 
 #define STRICT
 #include <windows.h>
@@ -785,7 +785,7 @@ init(void)
 	}
 	if (strlen(szLogFile) == 0) {
 		MessageBox(HWND_DESKTOP, "Usage: uninstgs logfile.txt", 
-			"Ghostscript Uninstall", MB_OK);
+			"GPL Ghostscript Uninstall", MB_OK);
 		return FALSE;
 	}
 	
