@@ -1,4 +1,4 @@
-/* Copyright (C) 2002 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 2002-2004 artofcode LLC. All rights reserved.
   
   This software is provided AS-IS with no warranty, either express or
   implied.
@@ -14,12 +14,14 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/* $Id: gdevpdte.c,v 1.20 2002/11/18 14:29:54 igor Exp $ */
+/* $Id: gdevpdte.c,v 1.33.2.12 2004/09/09 11:45:19 igor Exp $ */
 /* Encoding-based (Type 1/2/42) text processing for pdfwrite. */
+
 #include "math_.h"
 #include "memory_.h"
 #include "gx.h"
 #include "gserrors.h"
+#include "gsutil.h"
 #include "gxfcmap.h"
 #include "gxfcopy.h"
 #include "gxfont.h"
@@ -39,9 +41,9 @@ private int pdf_char_widths(gx_device_pdf *const pdev,
 			    pdf_font_resource_t *pdfont, int ch,
 			    gs_font_base *font,
 			    pdf_glyph_widths_t *pwidths /* may be NULL */);
-private int pdf_encode_string(gx_device_pdf *pdev, const pdf_text_enum_t *penum, 
-			      const gs_string *pstr,
-			      pdf_font_resource_t **ppdfont);
+private int pdf_encode_string(gx_device_pdf *pdev, const pdf_text_enum_t *penum,
+		  const gs_string *pstr, const gs_glyph *gdata,
+		  pdf_font_resource_t **ppdfont);
 private int pdf_process_string(pdf_text_enum_t *penum, gs_string *pstr,
 			       pdf_font_resource_t *pdfont,
 			       const gs_matrix *pfmat,
@@ -52,7 +54,7 @@ private int pdf_process_string(pdf_text_enum_t *penum, gs_string *pstr,
  */
 int
 pdf_encode_process_string(pdf_text_enum_t *penum, gs_string *pstr,
-			  const gs_matrix *pfmat,
+			  const gs_glyph *gdata, const gs_matrix *pfmat,
 			  pdf_text_process_state_t *ppts)
 {
     gx_device_pdf *const pdev = (gx_device_pdf *)penum->dev;
@@ -64,18 +66,105 @@ pdf_encode_process_string(pdf_text_enum_t *penum, gs_string *pstr,
     case ft_TrueType:
     case ft_encrypted:
     case ft_encrypted2:
+    case ft_user_defined:
 	break;
     default:
 	return_error(gs_error_rangecheck);
     }
     font = (gs_font_base *)penum->current_font;
 
-    code = pdf_encode_string(pdev, penum, pstr, &pdfont); /* Must not change penum. */
+    code = pdf_encode_string(pdev, penum, pstr, gdata, &pdfont); /* Must not change penum. */
     if (code < 0)
 	return code;
     return pdf_process_string(penum, pstr, pdfont, pfmat, ppts);
 }
 
+/* 
+ * Add char code pair to ToUnicode CMap,
+ * creating the CMap on neccessity.
+ */
+int
+pdf_add_ToUnicode(gx_device_pdf *pdev, gs_font *font, pdf_font_resource_t *pdfont, 
+		  gs_glyph glyph, gs_char ch)
+{   int code;
+    gs_char unicode;
+
+    if (glyph == GS_NO_GLYPH)
+	return 0;
+    unicode = font->procs.decode_glyph((gs_font *)font, glyph);
+    if (unicode != GS_NO_CHAR) {
+	if (pdfont->cmap_ToUnicode == NULL) {
+	    uint num_codes = 256, key_size = 1;
+	    
+	    if (font->FontType == ft_CID_encrypted) {
+		gs_font_cid0 *pfcid = (gs_font_cid0 *)font;
+
+		num_codes = pfcid->cidata.common.CIDCount;
+		key_size = 2;
+	    } else if (font->FontType == ft_CID_TrueType) {
+		gs_font_cid2 *pfcid = (gs_font_cid2 *)font;
+
+		num_codes = pfcid->cidata.common.CIDCount;
+		key_size = 2;
+	    }
+	    code = gs_cmap_ToUnicode_alloc(pdev->pdf_memory, pdfont->rid, num_codes, key_size, 
+					    &pdfont->cmap_ToUnicode);
+	    if (code < 0)
+		return code;
+	}
+	if (pdfont->cmap_ToUnicode != NULL)
+	    gs_cmap_ToUnicode_add_pair(pdfont->cmap_ToUnicode, ch, unicode);
+    }
+    return 0;
+}
+
+/*
+ * If the current substream is a charproc, register a font used in it.
+ */
+private int
+pdf_register_charproc_font(gx_device_pdf *pdev, gs_id id) 
+{
+    if (pdev->font3 != 0) {
+	pdf_font_resource_t *pdfont = (pdf_font_resource_t *)pdev->font3;
+	gs_id *used_fonts = pdfont->u.simple.s.type3.used_fonts;
+	int i, used_fonts_count = pdfont->u.simple.s.type3.used_fonts_count;
+
+	for (i = 0; i < used_fonts_count; i++)
+	    if (used_fonts[i] == id)
+		return 0;
+	if (used_fonts_count >= count_of(pdfont->u.simple.s.type3.used_fonts))
+	    return_error(gs_error_limitcheck);
+	used_fonts[used_fonts_count] = id;
+	pdfont->u.simple.s.type3.used_fonts_count = used_fonts_count + 1;
+    }
+    return 0;
+}
+
+/*
+ * Register charproc fonts with the page or substream.
+ */
+int
+pdf_used_charproc_fonts(gx_device_pdf *pdev, pdf_font_resource_t *pdfont) 
+{
+    if (pdfont->where_used & pdev->used_mask)
+	return 0;
+    pdfont->where_used |= pdev->used_mask;
+    if (pdfont->FontType == ft_user_defined) {
+	gs_id *used_fonts = pdfont->u.simple.s.type3.used_fonts;
+	int i, used_fonts_count = pdfont->u.simple.s.type3.used_fonts_count;
+
+	for (i = 0; i < used_fonts_count; i++) {
+	    pdf_font_resource_t *pdfont1 = 
+		    (pdf_font_resource_t *)pdf_find_resource_by_gs_id(pdev, 
+			    resourceFont, used_fonts[i]);
+	    if (pdfont1 == NULL)
+		return_error(gs_error_unregistered); /* Must not happen. */
+	    pdfont1->where_used |= pdev->used_mask;
+	}
+    }
+    return 0;
+}
+
 /*
  * Given a text string and a simple gs_font, return a font resource suitable
  * for the text string, possibly re-encoding the string.  This
@@ -86,27 +175,36 @@ pdf_encode_process_string(pdf_text_enum_t *penum, gs_string *pstr,
  */
 private int
 pdf_encode_string(gx_device_pdf *pdev, const pdf_text_enum_t *penum,
-		  const gs_string *pstr,
+		  const gs_string *pstr, const gs_glyph *gdata,
 		  pdf_font_resource_t **ppdfont)
 {
     gs_font *font = (gs_font *)penum->current_font;
     pdf_font_resource_t *pdfont = 0;
-    gs_font_base *cfont;
+    gs_font_base *cfont, *ccfont;
     int code, i;
 
     /*
-     * This crude version of the code simply uses pdf_obtain_font_resource,
-     * and never re-encodes characters.
+     * In contradiction with pre-7.20 versions of pdfwrite,
+     * we never re-encode texts due to possible encoding conflict while font merging.
      */
     code = pdf_obtain_font_resource((const gs_text_enum_t *)penum, pstr, &pdfont);
     if (code < 0)
 	return code;
-    cfont = pdf_font_resource_font(pdfont);
+    code = pdf_add_resource(pdev, pdev->substream_Resources, "/Font", (pdf_resource_t *)pdfont);
+    if (code < 0)
+	return code;
+    code = pdf_register_charproc_font(pdev, pdfont->rid);
+    if (code < 0)
+	return code;
+    cfont = pdf_font_resource_font(pdfont, false);
+    ccfont = pdf_font_resource_font(pdfont, true);
     for (i = 0; i < pstr->size; ++i) {
 	int ch = pstr->data[i];
 	pdf_encoding_element_t *pet = &pdfont->u.simple.Encoding[ch];
-	gs_glyph glyph =
-	    font->procs.encode_char(font, ch, GLYPH_SPACE_NAME);
+	gs_glyph glyph = (gdata == NULL
+			    ? font->procs.encode_char(font, ch, GLYPH_SPACE_NAME)
+			    : gdata[i]);
+		
 	gs_glyph copied_glyph;
 	gs_const_string gnstr;
 
@@ -121,28 +219,63 @@ pdf_encode_string(gx_device_pdf *pdev, const pdf_text_enum_t *penum,
 	code = font->procs.glyph_name(font, glyph, &gnstr);
 	if (code < 0)
 	    return code;	/* can't get name of glyph */
-	/* The standard 14 fonts don't have a FontDescriptor. */
-	code = (pdfont->base_font != 0 ?
-		pdf_base_font_copy_glyph(pdfont->base_font, glyph, (gs_font_base *)font) :
-		pdf_font_used_glyph(pdfont->FontDescriptor, glyph, (gs_font_base *)font));
-	if (code == gs_error_undefined)
-	    continue;	/* notdef */
+	if (font->FontType != ft_user_defined) {
+	    /* The standard 14 fonts don't have a FontDescriptor. */
+	    code = (pdfont->base_font != 0 ?
+		    pdf_base_font_copy_glyph(pdfont->base_font, glyph, (gs_font_base *)font) :
+		    pdf_font_used_glyph(pdfont->FontDescriptor, glyph, (gs_font_base *)font));
+	    if (code < 0 && code != gs_error_undefined)
+		return code;
+	    if (code == gs_error_undefined) {
+		/* PS font has no such glyph. */
+		if (bytes_compare(gnstr.data, gnstr.size, (const byte *)".notdef", 7)) {
+		    pet->glyph = glyph;
+		    pet->str = gnstr;
+		    pet->is_difference = true;
+		}
+	    } else if (pdfont->base_font == NULL && ccfont != NULL &&
+		    (gs_copy_glyph_options(font, glyph, (gs_font *)ccfont, COPY_GLYPH_NO_NEW) != 1 || 
+		     gs_copied_font_add_encoding((gs_font *)ccfont, ch, glyph) < 0)) {
+		/*
+		 * The "complete" copy of the font appears incomplete
+		 * due to incrementally added glyphs. Drop the "complete"
+		 * copy now and continue with subset font only.
+		 *
+		 * Note that we need to add the glyph to the encoding of the
+		 * "complete" font, because "PPI-ProPag 2.6.1.4 (archivePg)"
+		 * creates multiple font copies with reduced encodings
+		 * (we believe it is poorly designed),
+		 * and we can merge the copies back to a single font (see Bug 686875).
+		 * We also check whether the encoding is compatible. 
+		 * It must be compatible here due to the pdf_obtain_font_resource 
+		 * and ccfont logics, but we want to ensure for safety reason.
+		 */
+		ccfont = NULL;
+		pdf_font_descriptor_drop_complete_font(pdfont->FontDescriptor);
+	    }
+	    /*
+	     * We arbitrarily allow the first encoded character in a given
+	     * position to determine the encoding associated with the copied
+	     * font.
+	     */
+	    copied_glyph = cfont->procs.encode_char((gs_font *)cfont, ch,
+						    GLYPH_SPACE_NAME);
+	    if (glyph != copied_glyph &&
+		gs_copied_font_add_encoding((gs_font *)cfont, ch, glyph) < 0
+		)
+		pet->is_difference = true;
+	    pdfont->used[ch >> 3] |= 0x80 >> (ch & 7);
+	} 
+	/*
+	 * We always generate ToUnicode for simple fonts, because
+	 * we can't detemine in advance, which glyphs the font actually uses.
+	 * The decision about writing it out is deferred until pdf_write_font_resource.
+	 */
+	code = pdf_add_ToUnicode(pdev, font, pdfont, glyph, ch);
 	if (code < 0)
 	    return code;
 	pet->glyph = glyph;
 	pet->str = gnstr;
-	/*
-	 * We arbitrarily allow the first encoded character in a given
-	 * position to determine the encoding associated with the copied
-	 * font.
-	 */
-	copied_glyph = cfont->procs.encode_char((gs_font *)cfont, ch,
-						GLYPH_SPACE_NAME);
-	if (glyph != copied_glyph &&
-	    gs_copied_font_add_encoding((gs_font *)cfont, ch, glyph) < 0
-	    )
-	    pet->is_difference = true;
-	pdfont->used[ch >> 3] |= 0x80 >> (ch & 7);
     }
     *ppdfont = pdfont;
     return 0;
@@ -152,7 +285,7 @@ pdf_encode_string(gx_device_pdf *pdev, const pdf_text_enum_t *penum,
  * Estimate text bbox.
  */
 private int
-process_text_estimate_bbox(gs_text_enum_t *pte, gs_font_base *font,
+process_text_estimate_bbox(pdf_text_enum_t *pte, gs_font_base *font,
 			  const gs_const_string *pstr,
 			  const gs_matrix *pfmat,
 			  gs_rect *text_bbox, gs_point *pdpt)
@@ -180,7 +313,7 @@ process_text_estimate_bbox(gs_text_enum_t *pte, gs_font_base *font,
     gs_matrix_multiply(pfmat, &m, &m);
     for (i = 0; i < pstr->size; ++i) {
 	byte c = pstr->data[i];
-	gs_rect bbox, bbox1;
+	gs_rect bbox;
 	gs_point wanted, tpt, p0, p1, p2, p3;
 	gs_glyph glyph = font->procs.encode_char((gs_font *)font, c, 
 					GLYPH_SPACE_NAME);
@@ -195,14 +328,14 @@ process_text_estimate_bbox(gs_text_enum_t *pte, gs_font_base *font,
 	gs_point_transform(font->FontBBox.p.x, font->FontBBox.q.y, &m, &p1);
 	gs_point_transform(font->FontBBox.q.x, font->FontBBox.p.y, &m, &p2);
 	gs_point_transform(font->FontBBox.q.x, font->FontBBox.q.y, &m, &p3);
-	bbox1.p.x = min(min(p0.x, p1.x), min(p1.x, p2.x)) + total.x;
-	bbox1.p.y = min(min(p0.y, p1.y), min(p1.y, p2.y)) + total.y;
-	bbox1.q.x = max(max(p0.x, p1.x), max(p1.x, p2.x)) + total.x;
-	bbox1.q.y = max(max(p0.y, p1.y), max(p1.y, p2.y)) + total.y;
+	bbox.p.x = min(min(p0.x, p1.x), min(p1.x, p2.x)) + total.x;
+	bbox.p.y = min(min(p0.y, p1.y), min(p1.y, p2.y)) + total.y;
+	bbox.q.x = max(max(p0.x, p1.x), max(p1.x, p2.x)) + total.x;
+	bbox.q.y = max(max(p0.y, p1.y), max(p1.y, p2.y)) + total.y;
 	if (i == 0)
-	    *text_bbox = bbox1;
+	    *text_bbox = bbox;
 	else
-	    rect_merge(*text_bbox, bbox1);
+	    rect_merge(*text_bbox, bbox);
 	if (pte->text.operation & TEXT_REPLACE_WIDTHS) {
 	    gs_text_replaced_width(&pte->text, xy_index++, &tpt);
 	    gs_distance_transform(tpt.x, tpt.y, &ctm_only(pte->pis), &wanted);
@@ -217,7 +350,7 @@ process_text_estimate_bbox(gs_text_enum_t *pte, gs_font_base *font,
 		wanted.x += tpt.x;
 		wanted.y += tpt.y;
 	    }
-	    if (pstr->data[i] == space_char) {
+	    if (pstr->data[i] == space_char && pte->text.operation & TEXT_ADD_TO_SPACE_WIDTH) {
 		gs_distance_transform(pte->text.delta_space.x,
 				      pte->text.delta_space.y,
 				      &ctm_only(pte->pis), &tpt);
@@ -232,6 +365,21 @@ process_text_estimate_bbox(gs_text_enum_t *pte, gs_font_base *font,
     return 0;
 }
 
+private void
+adjust_first_last_char(pdf_font_resource_t *pdfont, byte *str, int size)
+{
+    int i;
+
+    for (i = 0; i < size; ++i) {
+	int chr = str[i];
+
+	if (chr < pdfont->u.simple.FirstChar)
+	    pdfont->u.simple.FirstChar = chr;
+	if (chr > pdfont->u.simple.LastChar)
+	    pdfont->u.simple.LastChar = chr;
+    }
+}
+
 /*
  * Internal procedure to process a string in a non-composite font.
  * Doesn't use or set pte->{data,size,index}; may use/set pte->xy_index;
@@ -242,56 +390,58 @@ process_text_estimate_bbox(gs_text_enum_t *pte, gs_font_base *font,
  * necessary glyphs.  penum->current_font provides the gs_font for getting
  * glyph metrics, but this font's Encoding is not used.
  */
-private int process_text_return_width(const gs_text_enum_t *pte,
+private int process_text_return_width(const pdf_text_enum_t *pte,
 				      gs_font_base *font,
 				      pdf_text_process_state_t *ppts,
 				      const gs_const_string *pstr,
-				      gs_point *pdpt);
-private int process_text_modify_width(gs_text_enum_t *pte,
-				      gs_font_base *font,
-				      pdf_text_process_state_t *ppts,
-				      const gs_const_string *pstr,
-				      gs_point *pdpt);
+				      gs_point *pdpt, int *accepted);
 private int
 pdf_process_string(pdf_text_enum_t *penum, gs_string *pstr,
 		   pdf_font_resource_t *pdfont, const gs_matrix *pfmat,
 		   pdf_text_process_state_t *ppts)
 {
-    gs_text_enum_t *const pte = (gs_text_enum_t *)penum;
     gx_device_pdf *const pdev = (gx_device_pdf *)penum->dev;
     gs_font_base *font = (gs_font_base *)penum->current_font;
-    const gs_text_params_t *text = &pte->text;
-    int i;
+    const gs_text_params_t *text = &penum->text;
     int code = 0, mask;
     gs_point width_pt;
     gs_rect text_bbox;
+    int accepted;
 
     if (pfmat == 0)
 	pfmat = &font->FontMatrix;
     if (text->operation & TEXT_RETURN_WIDTH) {
-	code = gx_path_current_point(pte->path, &penum->origin);
+	code = gx_path_current_point(penum->path, &penum->origin);
 	if (code < 0)
 	    return code;
     }
-    
-    /*
-     * Acrobat Reader can't handle text with huge coordinates,
-     * so skip the text if it is outside the clip bbox.
-     */
-    code = process_text_estimate_bbox(pte, font, (gs_const_string *)pstr, pfmat, 
-				      &text_bbox, &width_pt);
-    if (code == 0) {
-	gs_fixed_rect clip_bbox;
-	gs_rect rect;
+    if (text->size == 0)
+	return 0;
+    if (penum->pis->text_rendering_mode != 3 && !(text->operation & TEXT_DO_NONE)) {
+	/*
+	 * Acrobat Reader can't handle text with huge coordinates,
+	 * so skip the text if it is outside the clip bbox
+	 * (Note : it ever fails with type 3 fonts).
+	 */
+	code = process_text_estimate_bbox(penum, font, (gs_const_string *)pstr, pfmat, 
+					  &text_bbox, &width_pt);
+	if (code == 0) {
+	    gs_fixed_rect clip_bbox;
+	    gs_rect rect;
 
-	gx_cpath_outer_box(pte->pcpath, &clip_bbox);
-	rect.p.x = fixed2float(clip_bbox.p.x);
-	rect.p.y = fixed2float(clip_bbox.p.y);
-	rect.q.x = fixed2float(clip_bbox.q.x);
-	rect.q.y = fixed2float(clip_bbox.q.y);
-	rect_intersect(rect, text_bbox);
-	if (rect.p.x > rect.q.x || rect.p.y > rect.q.y)
-	    goto finish;
+	    gx_cpath_outer_box(penum->pcpath, &clip_bbox);
+	    rect.p.x = fixed2float(clip_bbox.p.x);
+	    rect.p.y = fixed2float(clip_bbox.p.y);
+	    rect.q.x = fixed2float(clip_bbox.q.x);
+	    rect.q.y = fixed2float(clip_bbox.q.y);
+	    rect_intersect(rect, text_bbox);
+	    if (rect.p.x > rect.q.x || rect.p.y > rect.q.y) {
+		penum->index += pstr->size;
+		goto finish;
+	    }
+	}
+    } else {
+	/* We have no penum->pcpath. */
     }
 
     /*
@@ -302,7 +452,7 @@ pdf_process_string(pdf_text_enum_t *penum, gs_string *pstr,
     if (code > 0) {
 	/* Try not to emulate ADD_TO_WIDTH if we don't have to. */
 	if (code & TEXT_ADD_TO_SPACE_WIDTH) {
-	    if (!memchr(pstr->data, pte->text.space.s_char, pstr->size))
+	    if (!memchr(pstr->data, penum->text.space.s_char, pstr->size))
 		code &= ~TEXT_ADD_TO_SPACE_WIDTH;
 	}
     }
@@ -310,20 +460,6 @@ pdf_process_string(pdf_text_enum_t *penum, gs_string *pstr,
 	return code;
     mask = code;
 
-    /*
-     * For incrementally loaded fonts, expand FirstChar..LastChar
-     * if needed.
-     */
-
-    for (i = 0; i < pstr->size; ++i) {
-	int chr = pstr->data[i];
-
-	if (chr < pdfont->u.simple.FirstChar)
-	    pdfont->u.simple.FirstChar = chr;
-	if (chr > pdfont->u.simple.LastChar)
-	    pdfont->u.simple.LastChar = chr;
-    }
-
     if (text->operation & TEXT_REPLACE_WIDTHS)
 	mask |= TEXT_REPLACE_WIDTHS;
 
@@ -339,39 +475,70 @@ pdf_process_string(pdf_text_enum_t *penum, gs_string *pstr,
 	 */
 	if (!(text->operation & (TEXT_DO_DRAW | TEXT_RETURN_WIDTH)))
 	    return 0;
-	code = process_text_return_width(pte, font, ppts,
+	code = process_text_return_width(penum, font, ppts,
 					 (gs_const_string *)pstr,
-					 &width_pt);
+					 &width_pt, &accepted);
 	if (code < 0)
 	    return code;
 	if (code == 0) {
 	    /* No characters with redefined widths -- the fast case. */
-	    if (text->operation & TEXT_DO_DRAW) {
-		code = pdf_append_chars(pdev, pstr->data, pstr->size,
-					width_pt.x, width_pt.y);
+	    if (text->operation & TEXT_DO_DRAW || penum->pis->text_rendering_mode == 3) {
+		code = pdf_append_chars(pdev, pstr->data, accepted,
+					width_pt.x, width_pt.y, false);
 		if (code < 0)
 		    return code;
-	    }
+		adjust_first_last_char(pdfont, pstr->data, accepted);
+		penum->index += accepted;
+	    } else if (text->operation & TEXT_DO_NONE)
+		penum->index += accepted;
 	} else {
 	    /* Use the slow case.  Set mask to any non-zero value. */
 	    mask = TEXT_RETURN_WIDTH;
 	}
     }
     if (mask) {
-	code = process_text_modify_width(pte, font, ppts,
+	/* process_text_modify_width destroys text parameters, save them now. */
+        int index0 = penum->index, xy_index = penum->xy_index;
+	gs_text_params_t text = penum->text;
+	int xy_index_step = (penum->text.x_widths != NULL && /* see gs_text_replaced_width */
+			     penum->text.x_widths == penum->text.y_widths ? 2 : 1);
+	
+	if (penum->text.x_widths != NULL) {
+	    penum->text.x_widths += xy_index * xy_index_step;
+	}
+	if (penum->text.y_widths != NULL)
+	    penum->text.y_widths += xy_index * xy_index_step;
+	penum->xy_index = 0;
+	code = process_text_modify_width(penum, (gs_font *)font, ppts,
 					 (gs_const_string *)pstr,
 					 &width_pt);
+	if (penum->text.x_widths != NULL)
+	    penum->text.x_widths -= xy_index * xy_index_step;
+	if (penum->text.y_widths != NULL)
+	    penum->text.y_widths -= xy_index * xy_index_step;
+	penum->xy_index += xy_index;
+	adjust_first_last_char(pdfont, pstr->data, penum->index);
+	penum->text = text;
+	penum->index += index0;
 	if (code < 0)
 	    return code;
     }
 
 
+finish:
     /* Finally, return the total width if requested. */
     if (!(text->operation & TEXT_RETURN_WIDTH))
 	return 0;
-finish:
-    pte->returned.total_width = width_pt;
-    return gx_path_add_point(pte->path,
+    if (text->operation & TEXT_DO_NONE) {
+	/* stringwidth needs to transform to user space. */
+	gs_point p;
+
+	gs_distance_transform_inverse(width_pt.x, width_pt.y, &ctm_only(penum->pis), &p);
+	penum->returned.total_width.x += p.x;
+	penum->returned.total_width.y += p.y;
+    } else
+	penum->returned.total_width = width_pt;
+    return gx_path_add_point(penum->path,
 			     penum->origin.x + float2fixed(width_pt.x),
 			     penum->origin.y + float2fixed(width_pt.y));
 }
@@ -390,29 +557,29 @@ pdf_char_widths(gx_device_pdf *const pdev,
     int code;
     byte *glyph_usage;
     double *real_widths;
-    int char_cache_size;
+    int char_cache_size, width_cache_size;
     pdf_font_resource_t *pdfont1;
 
-    pdf_attached_font_resource(pdev, (gs_font *)font, &pdfont1, 
-				&glyph_usage, &real_widths, &char_cache_size);
+    code = pdf_attached_font_resource(pdev, (gs_font *)font, &pdfont1, 
+				&glyph_usage, &real_widths, &char_cache_size, &width_cache_size);
+    if (code < 0)
+	return code;
     if (pdfont1 != pdfont)
 	return_error(gs_error_unregistered); /* Must not happen. */
     if (ch < 0 || ch > 255)
 	return_error(gs_error_rangecheck);
-    if (ch > char_cache_size)
+    if (ch >= width_cache_size)
 	return_error(gs_error_unregistered); /* Must not happen. */
     if (pwidths == 0)
 	pwidths = &widths;
-    if (real_widths[ch] == 0) {
+    if (font->FontType != ft_user_defined && real_widths[ch] == 0) {
 	/* Might be an unused char, or just not cached. */
 	gs_glyph glyph = pdfont->u.simple.Encoding[ch].glyph;
 
-	code = pdf_glyph_widths(pdfont, glyph, font, pwidths);
+	code = pdf_glyph_widths(pdfont, font->WMode, glyph, (gs_font *)font, pwidths, NULL);
 	if (code < 0)
 	    return code;
-	pdfont->u.simple.v[ch] = pwidths->v;
-	if (font->WMode != 0 && code > 0 &&
-	    pwidths->v.x == 0 && pwidths->v.y == 0) {
+	if (font->WMode != 0 && code > 0 && !pwidths->replaced_v) {
 	    /*
 	     * The font has no Metrics2, so it must write
 	     * horizontally due to PS spec.
@@ -420,26 +587,44 @@ pdf_char_widths(gx_device_pdf *const pdev,
 	     * which is required by PDF spec.
 	     * Take it from WMode==0.
 	     */
-	    int save_WMode = font->WMode;
-	    font->WMode = 0; /* Temporary patch font because font->procs.glyph_info
-	                        has no WMode argument. */
-	    code = pdf_glyph_widths(pdfont, glyph, font, pwidths);
-	    font->WMode = save_WMode;
+	    code = pdf_glyph_widths(pdfont, 0, glyph, (gs_font *)font, pwidths, NULL);
 	}
+	if (pwidths->replaced_v) {
+	    pdfont->u.simple.v[ch].x = pwidths->real_width.v.x - pwidths->Width.v.x;
+	    pdfont->u.simple.v[ch].y = pwidths->real_width.v.y - pwidths->Width.v.y;
+	} else
+	    pdfont->u.simple.v[ch].x = pdfont->u.simple.v[ch].y = 0;
 	if (code == 0) {
 	    pdfont->Widths[ch] = pwidths->Width.w;
 	    real_widths[ch] = pwidths->real_width.w;
 	}
     } else {
+	if (font->FontType == ft_user_defined) {
+	    if (!(pdfont->used[ch >> 3] & 0x80 >> (ch & 7)))
+		return gs_error_undefined; /* The charproc was not accumulated. */
+	    if (!pdev->charproc_just_accumulated &&
+		!(pdfont->u.simple.s.type3.cached[ch >> 3] & 0x80 >> (ch & 7))) {
+		 /* The charproc uses setcharwidth. 
+		    Need to accumulate again to check for a glyph variation. */
+	    return gs_error_undefined;
+	    }
+	}
 	pwidths->Width.w = pdfont->Widths[ch];
-	pwidths->real_width.w = real_widths[ch];
-	pwidths->v = pdfont->u.simple.v[ch];
-	if (font->WMode) {
+	pwidths->Width.v = pdfont->u.simple.v[ch];
+	if (font->FontType == ft_user_defined) {
+	    pwidths->real_width.w = real_widths[ch * 2];
+	    pwidths->Width.xy.x = pwidths->Width.w;
+	    pwidths->Width.xy.y = 0;
+	    pwidths->real_width.xy.x = real_widths[ch * 2 + 0];
+	    pwidths->real_width.xy.y = real_widths[ch * 2 + 1];
+	} else if (font->WMode) {
+	    pwidths->real_width.w = real_widths[ch];
 	    pwidths->Width.xy.x = 0;
 	    pwidths->Width.xy.y = pwidths->Width.w;
 	    pwidths->real_width.xy.x = 0;
 	    pwidths->real_width.xy.y = pwidths->real_width.w;
 	} else {
+	    pwidths->real_width.w = real_widths[ch];
 	    pwidths->Width.xy.x = pwidths->Width.w;
 	    pwidths->Width.xy.y = 0;
 	    pwidths->real_width.xy.x = pwidths->real_width.w;
@@ -455,29 +640,49 @@ pdf_char_widths(gx_device_pdf *const pdev,
  * character had real_width != Width, otherwise 0.
  */
 private int
-process_text_return_width(const gs_text_enum_t *pte, gs_font_base *font,
+process_text_return_width(const pdf_text_enum_t *pte, gs_font_base *font,
 			  pdf_text_process_state_t *ppts,
 			  const gs_const_string *pstr,
-			  gs_point *pdpt)
+			  gs_point *pdpt, int *accepted)
 {
     int i;
     gs_point w;
-    double scale = 0.001 * ppts->values.size;
+    double scale;
     gs_point dpt;
     int num_spaces = 0;
     int space_char =
 	(pte->text.operation & TEXT_ADD_TO_SPACE_WIDTH ?
 	 pte->text.space.s_char : -1);
-    int widths_differ = 0;
+    int widths_differ = 0, code;
+	gx_device_pdf *pdev = (gx_device_pdf *)pte->dev;
+    pdf_font_resource_t *pdfont;
 
+    code = pdf_attached_font_resource(pdev, (gs_font *)font, &pdfont, NULL, NULL, NULL, NULL);
+    if (code < 0)
+	return code;
+    if (font->FontType == ft_user_defined) {
+	pdf_font3_scale(pdev, (gs_font *)font, &scale);
+	scale *= ppts->values.size;
+    } else
+	scale = 0.001 * ppts->values.size;
     for (i = 0, w.x = w.y = 0; i < pstr->size; ++i) {
 	pdf_glyph_widths_t cw;
-	int code = pdf_char_widths((gx_device_pdf *)pte->dev,
-	                           ppts->values.pdfont, pstr->data[i], font,
-				   &cw);
+	gs_char ch = pstr->data[i];
 
-	if (code < 0)
+	if (font->FontType == ft_user_defined &&
+	    (i > 0 || !pdev->charproc_just_accumulated) &&
+	    !(pdfont->u.simple.s.type3.cached[ch >> 3] & (0x80 >> (ch & 7))))
+	    code = gs_error_undefined;
+	else 
+	code = pdf_char_widths((gx_device_pdf *)pte->dev,
+	                           ppts->values.pdfont, ch, font,
+				   &cw);
+	if (code < 0) {
+	    if (i)
+		break;
+	    *accepted = 0;
 	    return code;
+	}
 	w.x += cw.real_width.xy.x;
 	w.y += cw.real_width.xy.y;
 	if (cw.real_width.xy.x != cw.Width.xy.x ||
@@ -487,10 +692,11 @@ process_text_return_width(const gs_text_enum_t *pte, gs_font_base *font,
 	if (pstr->data[i] == space_char)
 	    ++num_spaces;
     }
+    *accepted = i;
     gs_distance_transform(w.x * scale, w.y * scale,
 			  &ppts->values.matrix, &dpt);
     if (pte->text.operation & TEXT_ADD_TO_ALL_WIDTHS) {
-	int num_chars = pstr->size;
+	int num_chars = *accepted;
 	gs_point tpt;
 
 	gs_distance_transform(pte->text.delta_all.x, pte->text.delta_all.y,
@@ -524,6 +730,7 @@ pdf_glyph_origin(pdf_font_resource_t *pdfont, int ch, int WMode, gs_point *p)
 	case ft_encrypted:
 	case ft_encrypted2:
 	case ft_TrueType:
+	case ft_user_defined:
 	    *p = pdfont->u.simple.v[ch];
 	    break;
 	default:
@@ -535,24 +742,38 @@ pdf_glyph_origin(pdf_font_resource_t *pdfont, int ch, int WMode, gs_point *p)
 /*
  * Emulate TEXT_ADD_TO_ALL_WIDTHS and/or TEXT_ADD_TO_SPACE_WIDTH,
  * and implement TEXT_REPLACE_WIDTHS if requested.
- * We know that the Tw and Tc values are zero.  Uses and updates
- * ppts->values.matrix; uses ppts->values.pdfont.
+ * Uses and updates ppts->values.matrix; uses ppts->values.pdfont.
+ * 
+ * Destroys the text parameters in *pte. 
+ * The caller must restore them.
  */
-private int
-process_text_modify_width(gs_text_enum_t *pte, gs_font_base *font,
+int
+process_text_modify_width(pdf_text_enum_t *pte, gs_font *font,
 			  pdf_text_process_state_t *ppts,
 			  const gs_const_string *pstr,
 			  gs_point *pdpt)
 {
     gx_device_pdf *const pdev = (gx_device_pdf *)pte->dev;
-    int i;
-    double scale = 0.001 * ppts->values.size;
+    double scale;
     int space_char =
 	(pte->text.operation & TEXT_ADD_TO_SPACE_WIDTH ?
 	 pte->text.space.s_char : -1);
     int code = 0;
     gs_point start, total;
+    bool composite = (ppts->values.pdfont->FontType == ft_composite);
 
+    if (font->FontType == ft_user_defined) {
+	gx_device_pdf *pdev = (gx_device_pdf *)pte->dev;
+
+	pdf_font3_scale(pdev, font, &scale);
+	scale *= ppts->values.size;
+    } else
+	scale = 0.001 * ppts->values.size;
+    pte->text.data.bytes = pstr->data;
+    pte->text.size = pstr->size;
+    pte->index = 0;
+    pte->text.operation &= ~TEXT_FROM_ANY;
+    pte->text.operation |= TEXT_FROM_STRING;
     start.x = ppts->values.matrix.tx;
     start.y = ppts->values.matrix.ty;
     total.x = total.y = 0;	/* user space */
@@ -561,21 +782,75 @@ process_text_modify_width(gs_text_enum_t *pte, gs_font_base *font,
      * values and the width value returned in *pdpt are in user space,
      * and the width values for pdf_append_chars are in device space.
      */
-    for (i = 0; i < pstr->size; ++i) {
+    for (;;) {
 	pdf_glyph_widths_t cw;	/* design space */
-	int code = pdf_char_widths((gx_device_pdf *)pte->dev,
-	                           ppts->values.pdfont, pstr->data[i], font,
-				   &cw);
 	gs_point did, wanted, tpt;	/* user space */
-	gs_point v; /* design space */
+	gs_point v = {0, 0}; /* design space */
+	gs_char chr;
+	gs_glyph glyph;
+	int code, index = pte->index;
+	gs_text_enum_t pte1 = *(gs_text_enum_t *)pte;
+	int FontType;
 
+	code = pte1.orig_font->procs.next_char_glyph(&pte1, &chr, &glyph);
+	if (code == 2) { /* end of string */
+	    gs_text_enum_copy_dynamic((gs_text_enum_t *)pte, &pte1, true);
+	    break;
+	}
 	if (code < 0)
 	    return code;
-	pdf_glyph_origin(ppts->values.pdfont, pstr->data[i], font->WMode, &v);
+	if (composite) { /* from process_cmap_text */
+	    gs_font *subfont = pte1.fstack.items[pte1.fstack.depth].font;
+	    pdf_font_resource_t *pdsubf = ppts->values.pdfont->u.type0.DescendantFont;
+
+	    FontType = pdsubf->FontType;
+	    code = pdf_glyph_widths(pdsubf, font->WMode, glyph, subfont, &cw, 
+		pte->cdevproc_callout ? pte->cdevproc_result : NULL);
+	} else {/* must be a base font */
+	    FontType = font->FontType;
+	    if (chr == GS_NO_CHAR && glyph != GS_NO_GLYPH) {
+		/* glyphshow, we have no char code. Bug 686988.*/
+		code = pdf_glyph_widths(ppts->values.pdfont, font->WMode, glyph, font, &cw, NULL);
+	    } else 
+		code = pdf_char_widths((gx_device_pdf *)pte->dev,
+				       ppts->values.pdfont, chr, (gs_font_base *)font,
+				       &cw);
+	}
+	if (code < 0) {
+	    if (index > 0)
+		break;
+	    return code;
+	}
+	gs_text_enum_copy_dynamic((gs_text_enum_t *)pte, &pte1, true);
+	if ((pte->text.operation & TEXT_FROM_SINGLE_GLYPH) ||
+	    (pte->text.operation & TEXT_FROM_GLYPHS)) {
+	    v.x = v.y = 0;
+	} else if (composite) {
+	    if (cw.replaced_v) {
+		v.x = cw.Width.v.x - cw.real_width.v.x;
+		v.y = cw.Width.v.y - cw.real_width.v.y;
+	    }
+	} else
+	    pdf_glyph_origin(ppts->values.pdfont, chr, font->WMode, &v);
 	if (v.x != 0 || v.y != 0) {
 	    gs_point glyph_origin_shift;
+	    double scale0;
 
-	    gs_distance_transform(-v.x * scale, -v.y * scale,
+	    if (FontType == ft_TrueType || FontType == ft_CID_TrueType)
+		scale0 = (float)0.001;
+	    else
+		scale0 = 1;
+	    glyph_origin_shift.x = -v.x * scale0;
+	    glyph_origin_shift.y = -v.y * scale0;
+	    gs_distance_transform(glyph_origin_shift.x, glyph_origin_shift.y,
+				  &font->FontMatrix, &glyph_origin_shift);
+	    if (composite) {
+		gs_font *subfont = pte->fstack.items[pte->fstack.depth].font;
+
+		gs_distance_transform(glyph_origin_shift.x, glyph_origin_shift.y,
+				      &subfont->FontMatrix, &glyph_origin_shift);
+	    }
+	    gs_distance_transform(glyph_origin_shift.x, glyph_origin_shift.y,
 				  &ctm_only(pte->pis), &glyph_origin_shift);
 	    if (glyph_origin_shift.x != 0 || glyph_origin_shift.y != 0) {
 		ppts->values.matrix.tx = start.x + total.x + glyph_origin_shift.x;
@@ -594,14 +869,14 @@ process_text_modify_width(gs_text_enum_t *pte, gs_font_base *font,
 				  &ppts->values.matrix, &tpt);
 	    did.x += tpt.x;
 	    did.y += tpt.y;
-	    if (pstr->data[i] == space_char) {
+	    if (chr == space_char) {
 		gs_distance_transform((font->WMode ? 0 : ppts->values.word_spacing),
 				      (font->WMode ? ppts->values.word_spacing : 0),
 				      &ppts->values.matrix, &tpt);
 		did.x += tpt.x;
 		did.y += tpt.y;
 	    }
-	    code = pdf_append_chars(pdev, &pstr->data[i], 1, did.x, did.y);
+	    code = pdf_append_chars(pdev, pstr->data + index, pte->index - index, did.x, did.y, composite);
 	    if (code < 0)
 		break;
 	} else
@@ -609,7 +884,9 @@ process_text_modify_width(gs_text_enum_t *pte, gs_font_base *font,
 	if (pte->text.operation & TEXT_REPLACE_WIDTHS) {
 	    gs_point dpt;
 
-	    gs_text_replaced_width(&pte->text, pte->xy_index++, &dpt);
+	    code = gs_text_replaced_width(&pte->text, pte->xy_index++, &dpt);
+	    if (code < 0)
+		return_error(gs_error_unregistered);
 	    gs_distance_transform(dpt.x, dpt.y, &ctm_only(pte->pis), &wanted);
 	} else {
 	    gs_distance_transform(cw.real_width.xy.x * scale,
@@ -622,7 +899,7 @@ process_text_modify_width(gs_text_enum_t *pte, gs_font_base *font,
 		wanted.x += tpt.x;
 		wanted.y += tpt.y;
 	    }
-	    if (pstr->data[i] == space_char) {
+	    if (chr == space_char && pte->text.operation & TEXT_ADD_TO_SPACE_WIDTH) {
 		gs_distance_transform(pte->text.delta_space.x,
 				      pte->text.delta_space.y,
 				      &ctm_only(pte->pis), &tpt);
@@ -639,12 +916,18 @@ process_text_modify_width(gs_text_enum_t *pte, gs_font_base *font,
 	    if (code < 0)
 		break;
 	}
+	pdev->charproc_just_accumulated = false;
     }
     *pdpt = total;
     return code;
 }
 
-private int 
+/*
+ * Get character code from a glyph code.
+ * An usage of this function is very undesirable,
+ * because a glyph may be unlisted in Encoding.
+ */
+int 
 pdf_encode_glyph(gs_font_base *bfont, gs_glyph glyph0,
 	    byte *buf, int buf_size, int *char_code_length)
 {
@@ -670,8 +953,7 @@ pdf_encode_glyph(gs_font_base *bfont, gs_glyph glyph0,
  * Process a text string in a simple font.
  */
 int
-process_plain_text(gs_text_enum_t *pte, const void *vdata, void *vbuf,
-		   uint size)
+process_plain_text(gs_text_enum_t *pte, void *vbuf, uint bsize)
 {
     byte *const buf = vbuf;
     uint count;
@@ -679,19 +961,28 @@ process_plain_text(gs_text_enum_t *pte, const void *vdata, void *vbuf,
     pdf_text_enum_t *const penum = (pdf_text_enum_t *)pte;
     int code;
     gs_string str;
-    bool encoded;
     pdf_text_process_state_t text_state;
+    const gs_glyph *gdata = NULL;
 
     if (operation & (TEXT_FROM_STRING | TEXT_FROM_BYTES)) {
-	count = size;
-	memcpy(buf, (const byte *)vdata + pte->index, size);
-	encoded = false;
+	count = pte->text.size - pte->index;
+	if (bsize < count)
+	    return_error(gs_error_unregistered); /* Must not happen. */
+	memcpy(buf, (const byte *)pte->text.data.bytes + pte->index, count);
     } else if (operation & (TEXT_FROM_CHARS | TEXT_FROM_SINGLE_CHAR)) {
 	/* Check that all chars fit in a single byte. */
-	const gs_char *const cdata = vdata;
+	const gs_char *cdata;
 	int i;
 
-	count = size / sizeof(gs_char);
+	if (operation & TEXT_FROM_CHARS) {
+	    cdata = pte->text.data.chars;
+	    count = (pte->text.size - pte->index);
+	} else {
+	    cdata = &pte->text.data.d_char;
+	    count = 1;
+	}
+	if (bsize < count * sizeof(gs_char))
+	    return_error(gs_error_unregistered); /* Must not happen. */
 	for (i = 0; i < count; ++i) {
 	    gs_char chr = cdata[pte->index + i];
 
@@ -699,14 +990,14 @@ process_plain_text(gs_text_enum_t *pte, const void *vdata, void *vbuf,
 		return_error(gs_error_rangecheck);
 	    buf[i] = (byte)chr;
 	}
-	encoded = false;
     } else if (operation & (TEXT_FROM_GLYPHS | TEXT_FROM_SINGLE_GLYPH)) {
 	/*
 	 * Since PDF has no analogue of 'glyphshow',
 	 * we try to encode glyphs with the current
 	 * font's encoding. If the current font has no encoding,
 	 * or the encoding doesn't contain necessary glyphs,
-	 * the text will be represented with a Type 3 font with bitmaps.
+	 * the text will be represented with a Type 3 font with 
+	 * bitmaps or outlines.
 	 *
 	 * When we fail with encoding (136-01.ps is an example),
 	 * we could locate a PDF font resource or create a new one
@@ -716,45 +1007,65 @@ process_plain_text(gs_text_enum_t *pte, const void *vdata, void *vbuf,
 	 * .notdef for a meanful printing).
 	 * fixme: Not implemented yet.
 	 */
-	const gs_glyph *const gdata = vdata;
 	gs_font *font = pte->current_font;
+	uint size;
 	int i;
 
+	if (operation & TEXT_FROM_GLYPHS) {
+	    gdata = pte->text.data.glyphs;
+	    size = pte->text.size - pte->index;
+	} else {
+	    gdata = &pte->text.data.d_glyph;
+	    size = 1;
+	}
 	if (!pdf_is_simple_font(font))
-	    return_error(gs_error_unregistered); /* Must not ahppen. */
+	    return_error(gs_error_unregistered); /* Must not happen. */
 	count = 0;
-	for (i = 0; i < size / sizeof(gs_glyph); ++i) {
+	for (i = 0; i < size; ++i) {
 	    gs_glyph glyph = gdata[pte->index + i];
 	    int char_code_length;
 
 	    code = pdf_encode_glyph((gs_font_base *)font, glyph, 
 			 buf + count, size - count, &char_code_length);
-	    if (code < 0) {
+	    if (code < 0)
+		break;
+	    count += char_code_length;
+	    if (operation & TEXT_INTERVENE)
+		break; /* Just do one character. */
+	}
+	if (i < size) {
+	    pdf_font_resource_t *pdfont; 
+
+	    str.data = buf;
+	    str.size = size;
+	    if (pdf_obtain_font_resource_unencoded(pte, &str, &pdfont, gdata) != 0) {
 		/* 
 		 * pdf_text_process will fall back 
 		 * to default implementation.
 		 */
 		return code;
 	    }
-	    count += char_code_length;
+	    count = size;
 	}
-	encoded = true;
+	/*  So far we will use TEXT_FROM_STRING instead
+	    TEXT_FROM_*_GLYPH*. Since we used a single 
+	    byte encoding, the character index appears invariant
+	    during this substitution.
+	 */
     } else
 	return_error(gs_error_rangecheck);
     str.data = buf;
-    if (count > 1 && (pte->text.operation & TEXT_INTERVENE)) {
+    if (count > 1 && (operation & TEXT_INTERVENE)) {
 	/* Just do one character. */
 	str.size = 1;
-	code = pdf_encode_process_string(penum, &str, NULL, &text_state);
+	code = pdf_encode_process_string(penum, &str, gdata, NULL, &text_state);
 	if (code >= 0) {
 	    pte->returned.current_char = buf[0];
 	    code = TEXT_PROCESS_INTERVENE;
 	}
     } else {
 	str.size = count;
-	code = pdf_encode_process_string(penum, &str, NULL, &text_state);
+	code = pdf_encode_process_string(penum, &str, gdata, NULL, &text_state);
     }
-    if (code >= 0)
-	pte->index += str.size;
     return code;
 }
