@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/* $Id: gstype42.c,v 1.25 2002/11/20 09:55:26 igor Exp $ */
+/* $Id: gstype42.c,v 1.30.2.3 2004/04/03 17:37:53 igor Exp $ */
 /* Type 42 (TrueType) font library routines */
 #include "memory_.h"
 #include "gx.h"
@@ -30,14 +30,23 @@
 #include "gxfont.h"
 #include "gxfont42.h"
 #include "gxttf.h"
+#include "gxttfb.h"
+#include "gxfcache.h"
 #include "gxistate.h"
+#include "stream.h"
 
 /* Structure descriptor */
 public_st_gs_font_type42();
 
 /* Forward references */
+#if NEW_TT_INTERPRETER
+private int append_outline_fitted(uint glyph_index, const gs_matrix * pmat,
+	       gx_path * ppath, cached_fm_pair * pair, 
+	       const gs_log2_scale_point * pscale, bool design_grid);
+#else
 private int append_outline(uint glyph_index, const gs_matrix_fixed * pmat,
 			   gx_path * ppath, gs_font_type42 * pfont);
+#endif
 private uint default_get_glyph_index(gs_font_type42 *pfont, gs_glyph glyph);
 private int default_get_outline(gs_font_type42 *pfont, uint glyph_index,
 				gs_glyph_data_t *pgd);
@@ -48,6 +57,7 @@ private int default_get_outline(gs_font_type42 *pfont, uint glyph_index,
   BEGIN\
     code = (*string_proc)(pfont, (ulong)(base), length, &vptr);\
     if ( code < 0 ) return code;\
+    if ( code > 0 ) return_error(gs_error_invalidfont);\
   END
 
 /* Get 2- or 4-byte quantities from a table. */
@@ -157,6 +167,9 @@ gs_type42_font_init(gs_font_type42 * pfont)
 	pfont->FontBBox.q.x = S16(head_box + 4) / upem;
 	pfont->FontBBox.q.y = S16(head_box + 6) / upem;
     }
+#if NEW_TT_INTERPRETER
+    pfont->data.warning_patented = false;
+#endif
     pfont->data.get_glyph_index = default_get_glyph_index;
     pfont->data.get_outline = default_get_outline;
     pfont->data.get_metrics = gs_type42_default_get_metrics;
@@ -335,9 +348,84 @@ default_get_outline(gs_font_type42 * pfont, uint glyph_index,
 	gs_glyph_data_from_null(pgd);
     else {
 	const byte *data;
+	byte *buf;
 
-	ACCESS(pfont->data.glyf + glyph_start, glyph_length, data);
-	gs_glyph_data_from_string(pgd, data, glyph_length, NULL);
+	code = (*string_proc)(pfont, (ulong)(pfont->data.glyf + glyph_start),
+			      glyph_length, &data);
+	if (code < 0) 
+	    return code;
+	if (code == 0)
+	    gs_glyph_data_from_string(pgd, data, glyph_length, NULL);
+	else {
+	    /*
+	     * The glyph is segmented in sfnts. 
+	     * It is not allowed with Type 42 specification.
+	     * Perhaps we can handle it (with a low performance),
+	     * making a contiguous copy.
+	     */
+	    uint left = glyph_length;
+
+	    /* 'code' is the returned length */
+	    buf = (byte *)gs_alloc_string(pfont->memory, glyph_length, "default_get_outline");
+	    if (buf == 0)
+		return_error(gs_error_VMerror);
+	    gs_glyph_data_from_string(pgd, buf, glyph_length, (gs_font *)pfont);
+	    for (;;) {
+		memcpy(buf + glyph_length - left, data, code);
+		if (!(left -= code))
+		    return 0;
+		code = (*string_proc)(pfont, (ulong)(pfont->data.glyf + glyph_start + 
+		              glyph_length - left), left, &data);
+		if (code < 0) 
+		    return code;
+		if (code == 0) 
+		    code = left;
+	    }
+	}
+    }
+    return 0;
+}
+
+/* Take outline data from a True Type font file. */
+int
+gs_type42_get_outline_from_TT_file(gs_font_type42 * pfont, stream *s, uint glyph_index,
+		gs_glyph_data_t *pgd)
+{
+    uchar lbuf[8];
+    ulong glyph_start;
+    uint glyph_length, count;
+
+    /* fixme: Since this function is being called multiple times per glyph,
+     * we should cache glyph data in a buffer associated with the font.
+     */
+    if (pfont->data.indexToLocFormat) {
+	sseek(s, pfont->data.loca + glyph_index * 4);
+        sgets(s, lbuf, 8, &count);
+	if (count < 8)
+	    return_error(gs_error_invalidfont);
+	glyph_start = u32(lbuf);
+	glyph_length = u32(lbuf + 4) - glyph_start;
+    } else {
+	sseek(s, pfont->data.loca + glyph_index * 2);
+        sgets(s, lbuf, 4, &count);
+	if (count < 4)
+	    return_error(gs_error_invalidfont);
+	glyph_start = (ulong) U16(lbuf) << 1;
+	glyph_length = ((ulong) U16(lbuf + 2) << 1) - glyph_start;
+    }
+    if (glyph_length == 0)
+	gs_glyph_data_from_null(pgd);
+    else {
+	byte *buf;
+
+	sseek(s, pfont->data.glyf + glyph_start);
+	buf = (byte *)gs_alloc_string(pfont->memory, glyph_length, "default_get_outline");
+	if (buf == 0)
+	    return_error(gs_error_VMerror);
+	gs_glyph_data_from_string(pgd, buf, glyph_length, (gs_font *)pfont);
+        sgets(s, buf, glyph_length, &count);
+	if (count < glyph_length)
+	    return_error(gs_error_invalidfont);
     }
     return 0;
 }
@@ -347,7 +435,9 @@ private int
 parse_pieces(gs_font_type42 *pfont, gs_glyph glyph, gs_glyph *pieces,
 	     int *pnum_pieces)
 {
-    uint glyph_index = pfont->data.get_glyph_index(pfont, glyph);
+    uint glyph_index = (glyph >= GS_MIN_GLYPH_INDEX 
+			? glyph - GS_MIN_GLYPH_INDEX 
+			: pfont->data.get_glyph_index(pfont, glyph));
     gs_glyph_data_t glyph_data;
     int code = pfont->data.get_outline(pfont, glyph_index, &glyph_data);
 
@@ -363,7 +453,7 @@ parse_pieces(gs_font_type42 *pfont, gs_glyph glyph, gs_glyph *pieces,
 	memset(&mat, 0, sizeof(mat)); /* arbitrary */
 	for (i = 0; flags & TT_CG_MORE_COMPONENTS; ++i) {
 	    if (pieces)
-		pieces[i] = U16(gdata + 2) + gs_min_cid_glyph;
+		pieces[i] = U16(gdata + 2) + GS_MIN_GLYPH_INDEX;
 	    parse_component(&gdata, &flags, &mat, NULL, pfont, &mat);
 	}
 	*pnum_pieces = i;
@@ -375,38 +465,65 @@ parse_pieces(gs_font_type42 *pfont, gs_glyph glyph, gs_glyph *pieces,
 
 /* Define the font procedures for a Type 42 font. */
 int
-gs_type42_glyph_outline(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
+gs_type42_glyph_outline(gs_font *font, int WMode, gs_glyph glyph, const gs_matrix *pmat,
 			gx_path *ppath)
 {
     gs_font_type42 *const pfont = (gs_font_type42 *)font;
-    uint glyph_index = pfont->data.get_glyph_index(pfont, glyph);
+    uint glyph_index = (glyph >= GS_MIN_GLYPH_INDEX 
+		? glyph - GS_MIN_GLYPH_INDEX 
+		: pfont->data.get_glyph_index(pfont, glyph));
     gs_fixed_point origin;
     int code;
     gs_glyph_info_t info;
+#if !NEW_TT_INTERPRETER
     gs_matrix_fixed fmat;
+#endif
     static const gs_matrix imat = { identity_matrix_body };
+#if NEW_TT_INTERPRETER
+    bool design_grid = true;
+    const gs_log2_scale_point log2_scale = {0, 0}; 
+    /* fixme : The subpixel numbers doesn't pass through the font_proc_glyph_outline interface.
+       High level devices can't get a proper grid fitting with AlignToPixels = 1.
+       Currently font_proc_glyph_outline is only used by pdfwrite for computing a
+       character bbox, which doesn't need a grid fitting.
+       We apply design grid here.
+     */
+    cached_fm_pair *pair;
+    code = gx_lookup_fm_pair(font, pmat, &log2_scale, design_grid, &pair);
+
+    if (code < 0)
+	return code;
+#endif
 
     if (pmat == 0)
 	pmat = &imat;
-    if ((code = gs_matrix_fixed_from_matrix(&fmat, pmat)) < 0 ||
-	(code = gx_path_current_point(ppath, &origin)) < 0 ||
+    if ((code = gx_path_current_point(ppath, &origin)) < 0 ||
+#if NEW_TT_INTERPRETER
+	(code = append_outline_fitted(glyph_index, pmat, ppath, pair, 
+					&log2_scale, design_grid)) < 0 ||
+#else
+	(code = gs_matrix_fixed_from_matrix(&fmat, pmat)) < 0 ||
 	(code = append_outline(glyph_index, &fmat, ppath, pfont)) < 0 ||
+	/* fixme : don't call glyph_info. */
+#endif
 	(code = font->procs.glyph_info(font, glyph, pmat,
-				       GLYPH_INFO_WIDTH, &info)) < 0
+				       GLYPH_INFO_WIDTH0 << WMode, &info)) < 0
 	)
 	return code;
-    return gx_path_add_point(ppath, origin.x + float2fixed(info.width[0].x),
-			     origin.y + float2fixed(info.width[0].y));
+    return gx_path_add_point(ppath, origin.x + float2fixed(info.width[WMode].x),
+			     origin.y + float2fixed(info.width[WMode].y));
 }
+
+/* Get glyph info by glyph index. */
 int
-gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
-		     int members, gs_glyph_info_t *info)
+gs_type42_glyph_info_by_gid(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
+		     int members, gs_glyph_info_t *info, uint glyph_index)
 {
     gs_font_type42 *const pfont = (gs_font_type42 *)font;
-    uint glyph_index = pfont->data.get_glyph_index(pfont, glyph);
     int default_members =
 	members & ~(GLYPH_INFO_WIDTHS | GLYPH_INFO_NUM_PIECES |
-		    GLYPH_INFO_PIECES | GLYPH_INFO_OUTLINE_WIDTHS);
+		    GLYPH_INFO_PIECES | GLYPH_INFO_OUTLINE_WIDTHS |
+		    GLYPH_INFO_VVECTOR0 | GLYPH_INFO_VVECTOR1);
     gs_glyph_data_t outline;
     int code = 0;
 
@@ -421,7 +538,7 @@ gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
 	gs_glyph_data_free(&outline, "gs_type42_glyph_info");
 	info->members = 0;
     }
-    if (members & GLYPH_INFO_WIDTH) {
+    if (members & GLYPH_INFO_WIDTHS) {
 	int i;
 
 	for (i = 0; i < 2; ++i)
@@ -429,8 +546,10 @@ gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
 		float sbw[4];
 
 		code = gs_type42_wmode_metrics(pfont, glyph_index, i, sbw);
-		if (code < 0)
-		    return code;
+		if (code < 0) {
+		    code = 0;
+		    continue;
+		}
 		if (pmat) {
 		    code = gs_point_transform(sbw[2], sbw[3], pmat,
 					      &info->width[i]);
@@ -442,8 +561,10 @@ gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
 		    info->width[i].x = sbw[2], info->width[i].y = sbw[3];
 		    info->v.x = sbw[0], info->v.y = sbw[1];
 		}
+		info->members |= (GLYPH_INFO_VVECTOR0 << i);
+		info->members |= (GLYPH_INFO_WIDTH << i);
 	    }
-	info->members |= members & GLYPH_INFO_WIDTH;
+	
     }
     if (members & (GLYPH_INFO_NUM_PIECES | GLYPH_INFO_PIECES)) {
 	gs_glyph *pieces =
@@ -457,6 +578,23 @@ gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
     return code;
 }
 int
+gs_type42_glyph_info(gs_font *font, gs_glyph glyph, const gs_matrix *pmat,
+		     int members, gs_glyph_info_t *info)
+{
+    gs_font_type42 *const pfont = (gs_font_type42 *)font;
+    uint glyph_index;
+    
+    if (glyph >= GS_MIN_GLYPH_INDEX)
+	glyph_index = glyph - GS_MIN_GLYPH_INDEX;
+    else {
+	glyph_index = pfont->data.get_glyph_index(pfont, glyph);
+	if (glyph_index == GS_NO_GLYPH)
+	    return_error(gs_error_undefined);
+    }
+    return gs_type42_glyph_info_by_gid(font, glyph, pmat, members, info, glyph_index);
+
+}
+int
 gs_type42_enumerate_glyph(gs_font *font, int *pindex,
 			  gs_glyph_space_t glyph_space, gs_glyph *pglyph)
 {
@@ -471,7 +609,7 @@ gs_type42_enumerate_glyph(gs_font *font, int *pindex,
 	    return code;
 	if (outline.bits.data == 0)
 	    continue;		/* empty (undefined) glyph */
-	*pglyph = glyph_index + gs_min_cid_glyph;
+	*pglyph = glyph_index + GS_MIN_GLYPH_INDEX;
 	gs_glyph_data_free(&outline, "gs_type42_enumerate_glyph");
 	return 0;
     }
@@ -497,6 +635,8 @@ simple_glyph_metrics(gs_font_type42 * pfont, uint glyph_index, int wmode,
 	uint num_metrics = pmtx->numMetrics;
 	const byte *pmetrics;
 
+	if (pmtx->length == 0)
+	    return_error(gs_error_rangecheck);
 	if (glyph_index < num_metrics) {
 	    ACCESS(pmtx->offset + glyph_index * 4, 4, pmetrics);
 	    width = U16(pmetrics);
@@ -584,6 +724,22 @@ gs_type42_get_metrics(gs_font_type42 * pfont, uint glyph_index,
 
 /* Append a TrueType outline to a path. */
 /* Note that this does not append the final moveto for the width. */
+#if NEW_TT_INTERPRETER
+int
+gs_type42_append(uint glyph_index, gs_imager_state * pis,
+		 gx_path * ppath, const gs_log2_scale_point * pscale,
+		 bool charpath_flag, int paint_type, cached_fm_pair *pair)
+{
+    int code = append_outline_fitted(glyph_index, &ctm_only(pis), ppath, 
+			pair, pscale, charpath_flag);
+
+    if (code < 0)
+	return code;
+    /* Set the flatness for curve rendering. */
+    return gs_imager_setflat(pis, gs_char_flatness(pis, 1.0));
+}
+
+#else
 int
 gs_type42_append(uint glyph_index, gs_imager_state * pis,
 		 gx_path * ppath, const gs_log2_scale_point * pscale,
@@ -596,6 +752,8 @@ gs_type42_append(uint glyph_index, gs_imager_state * pis,
     /* Set the flatness for curve rendering. */
     return gs_imager_setflat(pis, gs_char_flatness(pis, 1.0));
 }
+#endif
+
 
 /* Add 2nd degree Bezier to the path */
 private int
@@ -892,6 +1050,8 @@ append_component(uint glyph_index, const gs_matrix_fixed * pmat,
     gs_glyph_data_free(&glyph_data, "append_component");
     return code;
 }
+
+#if !NEW_TT_INTERPRETER
 private int
 append_outline(uint glyph_index, const gs_matrix_fixed * pmat,
 	       gx_path * ppath, gs_font_type42 * pfont)
@@ -935,3 +1095,20 @@ append_outline(uint glyph_index, const gs_matrix_fixed * pmat,
     gs_glyph_data_free(&glyph_data, "append_outline");
     return code;
 }
+#else
+private int
+append_outline_fitted(uint glyph_index, const gs_matrix * pmat,
+	       gx_path * ppath, cached_fm_pair * pair, 
+	       const gs_log2_scale_point * pscale, bool design_grid)
+{
+    gs_font_type42 *pfont = (gs_font_type42 *)pair->font;
+    int code;
+
+    gx_ttfReader__set_font(pair->ttr, pfont);
+    code = gx_ttf_outline(pair->ttf, pair->ttr, pfont, (uint)glyph_index, 
+	pmat, pscale, ppath, design_grid);
+    gx_ttfReader__set_font(pair->ttr, NULL);
+    return code;
+}
+#endif
+
