@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/* $Id: gximage3.c,v 1.11 2002/10/25 00:27:10 alexcher Exp $ */
+/* $Id: gximage3.c,v 1.11.2.2 2004/08/31 04:03:30 ray Exp $ */
 /* ImageType 3 image implementation */
 #include "math_.h"		/* for ceil, floor */
 #include "memory_.h"
@@ -328,8 +328,9 @@ gx_begin_image3_generic(gx_device * dev,
 	(code = gs_bbox_transform(&mrect, &mat, &mrect)) < 0
 	)
 	return code;
-    origin.x = (int)floor(mrect.p.x);
-    origin.y = (int)floor(mrect.p.y);
+
+    origin.x = (mrect.p.x < 0) ? (int)ceil(mrect.p.x) : (int)floor(mrect.p.x);
+    origin.y = (mrect.p.y < 0) ? (int)ceil(mrect.p.y) : (int)floor(mrect.p.y);
     code = make_mid(&mdev, dev, (int)ceil(mrect.q.x) - origin.x,
 		    (int)ceil(mrect.q.y) - origin.y, mem);
     if (code < 0)
@@ -348,7 +349,7 @@ gx_begin_image3_generic(gx_device * dev,
 	gx_drawing_color dcolor;
 	gs_matrix m_mat;
 
-	color_set_pure(&dcolor, 1);
+	set_nonclient_dev_color(&dcolor, 1);
 	/*
 	 * Adjust the translation for rendering the mask to include a
 	 * negative translation by origin.{x,y} in device space.
