@@ -14,12 +14,12 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/* $Id: iscan.c,v 1.13 2002/11/01 23:43:09 ghostgum Exp $ */
+/* $Id: iscan.c,v 1.13.2.2 2004/08/31 04:03:30 ray Exp $ */
 /* Token scanner for Ghostscript interpreter */
 #include "ghost.h"
 #include "memory_.h"
 #include "stream.h"
-#include "errors.h"
+#include "ierrors.h"
 #include "btoken.h"		/* for ref_binary_object_format */
 #include "files.h"		/* for fptr */
 #include "ialloc.h"
@@ -47,15 +47,6 @@
 #define recognize_btokens()\
   (ref_binary_object_format.value.intval != 0 && level2_enabled)
 
-#ifdef DEBUG
-/* Dummy comment processing procedure for testing. */
-private int
-no_comment_proc(const byte * str, uint len)
-{
-    return 0;
-}
-#endif
-
 /* Procedure for handling DSC comments if desired. */
 /* Set at initialization if a DSC handling module is included. */
 int (*scan_dsc_proc) (const byte *, uint) = NULL;
@@ -443,6 +434,7 @@ scan_token(i_ctx_t *i_ctx_p, stream * s, ref * pref, scanner_state * pstate)
     int sign;
     const bool check_only = (pstate->s_options & SCAN_CHECK_ONLY) != 0;
     const bool PDFScanRules = (i_ctx_p->scanner_options & SCAN_PDF_RULES) != 0;
+    const bool PDFScanInvNum = (i_ctx_p->scanner_options & SCAN_PDF_INV_NUM) != 0;
     scanner_state sstate;
 
 #define pstack sstate.s_pstack
@@ -452,6 +444,7 @@ scan_token(i_ctx_t *i_ctx_p, stream * s, ref * pref, scanner_state * pstate)
 #define name_type sstate.s_ss.s_name.s_name_type
 #define try_number sstate.s_ss.s_name.s_try_number
 
+    sptr = endptr = NULL; /* Quiet compiler */
     if (pstate->s_pstack != 0) {
 	if_not_spush1()
 	    return retcode;
@@ -686,7 +679,7 @@ scan_token(i_ctx_t *i_ctx_p, stream * s, ref * pref, scanner_state * pstate)
 	case '/':
 	    ensure2(scanning_none);
 	    c = scan_getc();
-	    if (c == '/') {
+	    if (!PDFScanRules && (c == '/')) {
 		name_type = 2;
 		c = scan_getc();
 	    } else
@@ -842,7 +835,7 @@ scan_token(i_ctx_t *i_ctx_p, stream * s, ref * pref, scanner_state * pstate)
 	     */
 	    retcode = scan_number(sptr + (sign & 1),
 		    endptr /*(*endptr == char_CR ? endptr : endptr + 1) */ ,
-				  sign, myref, &newptr, PDFScanRules);
+				  sign, myref, &newptr, PDFScanInvNum);
 	    if (retcode == 1 && decoder[newptr[-1]] == ctype_space) {
 		sptr = newptr - 1;
 		if (*sptr == char_CR && sptr[1] == char_EOL)
@@ -1051,7 +1044,7 @@ scan_token(i_ctx_t *i_ctx_p, stream * s, ref * pref, scanner_state * pstate)
 		const byte *base = da.base;
 
 		scan_sign(sign, base);
-		retcode = scan_number(base, daptr, sign, myref, &newptr, PDFScanRules);
+		retcode = scan_number(base, daptr, sign, myref, &newptr, PDFScanInvNum);
 		if (retcode == 1) {
 		    ref_mark_new(myref);
 		    retcode = 0;
