@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: zchar.c,v 1.10 2002/08/26 23:00:07 dan Exp $ */
+/*$Id: zchar.c,v 1.10.2.2 2004/07/29 17:53:51 igor Exp $ */
 /* Character operators */
 #include "ghost.h"
 #include "oper.h"
@@ -25,6 +25,7 @@
 #include "gxmatrix.h"		/* for ifont.h */
 #include "gxdevice.h"		/* for gxfont.h */
 #include "gxfont.h"
+#include "gxfont42.h"
 #include "gzstate.h"
 #include "dstack.h"		/* for stack depth */
 #include "estack.h"
@@ -37,6 +38,7 @@
 #include "iname.h"
 #include "ipacked.h"
 #include "store.h"
+#include "zchar42.h"
 
 /* Forward references */
 private bool map_glyph_to_char(const ref *, const ref *, ref *);
@@ -573,6 +575,28 @@ op_show_continue_dispatch(i_ctx_t *i_ctx_p, int npop, int code)
 	    ++esp;		/* skip BuildChar or BuildGlyph proc */
 	    return o_push_estack;
 	}
+	case TEXT_PROCESS_CDEVPROC:
+	    {   gs_font *pfont = penum->current_font;
+		ref cnref;
+		op_proc_t cont = op_show_continue, exec_cont = 0;
+		gs_glyph glyph = penum->returned.current_glyph;
+		int code;
+    
+		pop(npop);
+		op = osp;
+		glyph_ref(glyph, &cnref);
+		if (pfont->FontType == ft_CID_TrueType) {
+		    gs_font_type42 *pfont42 = (gs_font_type42 *)pfont;
+		    uint glyph_index = pfont42->data.get_glyph_index(pfont42, glyph);
+
+		    code = zchar42_set_cache(i_ctx_p, (gs_font_base *)pfont42, 
+				    &cnref, glyph_index, cont, &exec_cont, false);
+		} else
+		    return_error(e_unregistered); /* Unimplemented. */
+		if (exec_cont != 0)
+		    return_error(e_unregistered); /* Must not happen. */
+		return code;
+	    }
 	default:		/* error */
 err:
 	    if (code >= 0)
@@ -709,7 +733,8 @@ op_show_restore(i_ctx_t *i_ctx_p, bool for_error)
 	if (count > saved_count)	/* if <, we're in trouble */
 	    ref_stack_pop(&o_stack, count - saved_count);
     }
-    if (SHOW_IS_STRINGWIDTH(penum)) {	/* stringwidth does an extra gsave */
+    if (SHOW_IS_STRINGWIDTH(penum) && igs->text_rendering_mode != 3) {	
+	/* stringwidth does an extra gsave */
 	--saved_level;
     }
     if (penum->text.operation & TEXT_REPLACE_WIDTHS) {
@@ -779,6 +804,7 @@ font_bbox_param(const ref * pfdict, double bbox[4])
 	    int i;
 	    int code;
 	    float dx, dy, ratio;
+	    const float max_ratio = 12; /* From the bug 687594. */
 
 	    for (i = 0; i < 4; i++) {
 		packed_get(pbe, rbe + i);
@@ -786,12 +812,11 @@ font_bbox_param(const ref * pfdict, double bbox[4])
 	    }
 	    if ((code = num_params(rbe + 3, 4, bbox)) < 0)
 		return code;
-	    /* Require "reasonable" values.  Thanks to Ray */
-	    /* Johnston for suggesting the following test. */
+ 	    /* Require "reasonable" values. */
 	    dx = bbox[2] - bbox[0];
 	    dy = bbox[3] - bbox[1];
 	    if (dx <= 0 || dy <= 0 ||
-		(ratio = dy / dx) < 0.125 || ratio > 8.0
+		(ratio = dy / dx) < 1 / max_ratio || ratio > max_ratio
 		)
 		bbox[0] = bbox[1] = bbox[2] = bbox[3] = 0.0;
 	}
